//========================================================================
//
// Page.cc
//
// Copyright 1996 Derek B. Noonburg
//
//========================================================================
//
// Ported to EPOC by Sander van der Wal
//
// $Log: Page.cpp $
// Revision 1.2  2000-09-17 13:38:21+02  svdwal
// Ported
//

#ifdef __GNUC__
#pragma implementation
#endif

#include "Page.h"

// --o PdfLib
#include "Array.h"
#include "Dict.h"
#include "XRef.h"

#ifndef PDF_PARSER_ONLY
#include "Gfx.h"
#endif

#include "Error.h"

// --o Pdf
#include "Pdf.rsg"
#include "PROFILE.H"

//------------------------------------------------------------------------
// PageAttrs
//------------------------------------------------------------------------

void PageAttrs::ConstructL(PageAttrs *attrs, Dict *dict) {
  RAutoObject obj1, obj2;
  double w, h;

  // get old/default values
  if (attrs) {
    x1 = attrs->x1;
    y1 = attrs->y1;
    x2 = attrs->x2;
    y2 = attrs->y2;
    cropX1 = attrs->cropX1;
    cropY1 = attrs->cropY1;
    cropX2 = attrs->cropX2;
    cropY2 = attrs->cropY2;
    rotate = attrs->rotate;
    attrs->resources.copyL(&resources);
  } else {
    // set default MediaBox to 8.5" x 11" -- this shouldn't be necessary
    // but some (non-compliant) PDF files don't specify a MediaBox
    x1 = 0;
    y1 = 0;
    x2 = 612;
    y2 = 792;
    cropX1 = cropY1 = cropX2 = cropY2 = 0;
    rotate = 0;
    resources.initNull();
  }

  // media box
  dict->lookupL("MediaBox", &obj1);
  if (obj1.isArray() && obj1.arrayGetLength() == 4) {
    obj1.arrayGetL(0, &obj2);
    if (obj2.isNum())
      x1 = obj2.getNum();
    obj2.free();
    obj1.arrayGetL(1, &obj2);
    if (obj2.isNum())
      y1 = obj2.getNum();
    obj2.free();
    obj1.arrayGetL(2, &obj2);
    if (obj2.isNum())
      x2 = obj2.getNum();
    obj2.free();
    obj1.arrayGetL(3, &obj2);
    if (obj2.isNum())
      y2 = obj2.getNum();
    obj2.free();
  }
  obj1.free();

  // crop box
  dict->lookupL("CropBox", &obj1);
  if (obj1.isArray() && obj1.arrayGetLength() == 4) {
    obj1.arrayGetL(0, &obj2);
    if (obj2.isNum())
      cropX1 = obj2.getNum();
    obj2.free();
    obj1.arrayGetL(1, &obj2);
    if (obj2.isNum())
      cropY1 = obj2.getNum();
    obj2.free();
    obj1.arrayGetL(2, &obj2);
    if (obj2.isNum())
      cropX2 = obj2.getNum();
    obj2.free();
    obj1.arrayGetL(3, &obj2);
    if (obj2.isNum())
      cropY2 = obj2.getNum();
    obj2.free();
  }
  obj1.free();

  // if the MediaBox is excessively larger than the CropBox,
  // just use the CropBox
  limitToCropBox = gFalse;
  w = 0.25 * (cropX2 - cropX1);
  h = 0.25 * (cropY2 - cropY1);
  if (cropX2 > cropX1 &&
      (cropX1 - x1 > w || x2 - cropX2 > w ||
       cropY1 - y1 > h || y2 - cropY2 > h)) {
    limitToCropBox = gTrue;
  }

  // rotate
  dict->lookupL("Rotate", &obj1);
  if (obj1.isInt())
    rotate = obj1.getInt();
  obj1.free();
  while (rotate < 0)
    rotate += 360;
  while (rotate >= 360)
    rotate -= 360;

  // resource dictionary
  dict->lookupL("Resources", &obj1);
  if (obj1.isDict()) {
    resources.free();
    obj1.copyL(&resources);
  }
  obj1.free();
}

PageAttrs::~PageAttrs() {
  resources.free();
}

//------------------------------------------------------------------------
// Page
//------------------------------------------------------------------------

void Page::ConstructL(Dict *pageDict) {

  ok = gTrue;

  // annotations
  pageDict->lookupNFL("Annots", &annots);
  if (!(annots.isRef() || annots.isArray() || annots.isNull())) {
    error(-1, R_PAGE_ANNOTATIONS_OBJECT__PAGE__D__IS_WRONG_TYPE___S_,
	  num, annots.getTypeName());
    annots.free();
    goto err2;
  }

  // contents
  pageDict->lookupNFL("Contents", &contents);
  if (!(contents.isRef() || contents.isArray() ||
	contents.isNull())) {
    error(-1, R_PAGE_CONTENTS_OBJECT__PAGE__D__IS_WRONG_TYPE___S_,
	  num, contents.getTypeName());
    contents.free();
    goto err1;
  }
  
  return;

 err2:
  annots.initNull();
 err1:
  contents.initNull();
  ok = gFalse;
}

Page::~Page() {

  delete attrs;
  annots.free();
  contents.free();
}


void Page::displayL(OutputDev *out, int dpi, int rotate) {
  
#ifndef PDF_PARSER_ONLY

  PROFILE_START(PAGE__DISPLAY);
 
  RAutoObject obj;

#ifndef __SYMBIAN32__
  if (printCommands) {
    printf("***** MediaBox = ll:%g,%g ur:%g,%g\n",
	   getX1(), getY1(), getX2(), getY2());
    if (isCropped()) {
      printf("***** CropBox = ll:%g,%g ur:%g,%g\n",
	     getCropX1(), getCropY1(), getCropX2(), getCropY2());
    }
    printf("***** Rotate = %d\n", attrs->getRotate());
  }
#endif

  rotate += getRotate();
  if (rotate >= 360)
    rotate -= 360;
  else if (rotate < 0)
    rotate += 360;

  Gfx* gfx = new(ELeave) Gfx();
  CleanupStack::PushL(gfx);
  gfx->ConstructL(out, num, attrs->getResourceDict(),
                  dpi, getX1(), getY1(), getX2(), getY2(), isCropped(),
                  getCropX1(), getCropY1(), getCropX2(), getCropY2(), rotate);
  contents.fetchL(&obj);
  if (!obj.isNull())
    gfx->displayL(&obj);
  
  obj.free();
  CleanupStack::PopAndDestroy(); // gfx;
  
  PROFILE_STOP(PAGE__DISPLAY);

#endif
 
}
