// GDIOutputDev.h
//
// Copyright 1999-2000 Sander van der Wal
//
// $Id: Gdioutputdev.h 1.3 2000-09-24 21:17:11+02 svdwal Exp svdwal $

#ifndef GDIOUTPUTDEV_H
#define GDIOUTPUTDEV_H

#ifdef __GNUC__
#pragma interface
#endif

// Parents
#include "OutputDev.h"

// GDI
#ifndef __GDI_H__
#include <gdi.h>
#endif

#include "gdiview.h"


//-----------------------------------------------------------
// RPath.
//
// A RRegion with some extra convenience operators, used to
// implement a clipping path
//-----------------------------------------------------------
class CPolygon;

class RPath: public RRegion
{
public:
  RPath() {}
  RPath(const RPath& aPath): RRegion(aPath) {}

  void Set(const TRect& aRect);

  void AddPolygonL(const CPolygon& aPolygon, CGraphicsContext::TFillRule aFillRule);
  void AddStringL(const TDesC& aString);
  
  TBool IsRect() const { return Count() == 1; }
  TRect Rect() const  { return this->operator[](0); } ;
  
  TBool Intersects(const TRect& aRect);
  void  Intersect(const TRect& aRect);
  void  Intersect(const TRegion& aRegion) { RRegion::Intersect(aRegion); }
};


//--------------------------------
// TGDIOutputState
//
// drawing state for CGDIOutputDev
//--------------------------------
class CGDITextView;
class CGDIOutputDev;

class TGDIOutputState
{
public:
  TGDIOutputState(CGDIOutputDev* aDev);
  void RestoreAndPop(CGDIOutputDev* aDev);

public:
  TGDIOutputState* iNext;

  CArrayPtrFlat<CPolygon>*    iClipPath;   // current clip path as polygons
  CGraphicsContext::TFillRule iFillRule;   // of iClipPath
  RPath                       iClipRegion;
  
  CGraphicsContext*           iGraphicsGc;
  TInt                        iFlatness;
  TRgb                        iFillColor;
  TRgb                        iStrokeColor;
  TSize                       iLineWidth;
  CGraphicsContext::TPenStyle iLineDash;
  
  CGraphicsContext* iTextGc;
  CGDITextView*     iTextView;
};


//------------------------------------------------------------------------
// CGDIOutputDev
//
// Common interfaces and implementaions for EPOC devices
//------------------------------------------------------------------------
class GfxPath;
class GfxSubpath;

class CGDIOutputDev: public OutputDev
{
public:
  CGDIOutputDev();
  CGDIOutputDev(TBool aRenderText, TBool aRenderLineDrawings, TBool aRenderImages);
  //void ConstructL();
  ~CGDIOutputDev();

// --o OutputDev
public:
  //----- get info about output device

  // Does this device use upside-down coordinates?
  // (Upside-down means (0,0) is the top left corner of the page.)
  GBool upsideDown() { return gTrue; };

  // End a page.
  void endPage();

  //----- update graphics state
  void updateAll(GfxState *state);
  void updateCropBox(GfxState* state, double cropX1, double cropY1, double cropX2, double cropY2);
  void updateCTM(GfxState *state, double m11, double m12,
                 double m21, double m22, double m31, double m32);
  void updateLineDash(GfxState *state);
  void updateFlatness(GfxState *state);
  void updateLineWidth(GfxState *state);
  void updateFillColor(GfxState *state);
  void updateStrokeColor(GfxState *state);
  
  //----- update text state
  void updateFont(GfxState *state);
  void updateTextMat(GfxState *state);
  void updateRender(GfxState *state);

  //----- path painting
  void fill(GfxState *state, GBool strokeToo = gFalse);
  void eoFill(GfxState *state, GBool strokeToo = gFalse);

  //----- path clipping
  void clip(GfxState *state);
  void eoClip(GfxState *state);

  //----- text drawing
  void beginString(GfxState *state, GString *s);
  void endString(GfxState *state);
  void drawChar(GfxState *state, double x, double y,
                                 double dx, double dy, 
                                 Guchar c);
  void drawString(GfxState *state, GString *s);

// --o CGDIOutputDev
protected:
  // Start a page.
  void InitPageL();
  void CheckTypefacesL();
  
  // create a text view
  virtual void CreateTextViewL(TInt aRotate) = 0;

  // draw the filled polygon
  void ClearPath();
  virtual void FillPathL(CGraphicsContext::TFillRule aFillRule) = 0;
  void DoClipL(GfxState* aState, CGraphicsContext::TFillRule aFillRule);
  void MakeClipRegion();
  virtual void SetTextClipPath();
  virtual void SetGraphicsClipPath();

  void ConvertPathL(GfxState* aState);
  CPolygon* GetSubpathL(GfxSubpath* aSubpath, GfxState* aState);

  void GetImagePlacementInfo(GfxState *aState, TRect& aDevRect, TBool& aRotate, TBool& aXFlip, TBool& aYFlip);
  
  void DoDrawImageL(Stream *aStr, TSize aImgSize, 
                    GfxImageColorMap *aColorMap,
                    TBool aRotate, TBool aXFlip, TBool aYFlip,
                    TBool aDitherImages,
                    const TRect& aDevRect,
                    CWsBitmap* aDstBitmap, TDisplayMode aDisplayMode,
                    const TRect& aClipRect);

  TInt FindPixel(const TGDIRgb& aColor, TDisplayMode aDisplayMode, TGDIRgb& aError);

protected:
  friend class TGDIOutputState;

  TBool iRenderText;
  TBool iRenderLineDrawings;
  TBool iRenderImages;
  TBool iClip;
  
  CGraphicsDevice*  iDev;
  TGDIFontNameSubst iFontNameSubst[4];
 
  TDisplayMode iDisplayMode;
  TSize        iPageSizeInTwips;

  // saved state
  TGDIOutputState* iSave;

  // special graphics state
  CArrayPtrFlat<CPolygon>* iPath;  // current path
  GfxPath*                 iGfxPath; // gfx form of iPath

  // clipping path
  CArrayPtrFlat<CPolygon>* iClipPath;   // current clip path as polygons
  CGraphicsContext::TFillRule iFillRule; // of iClipPath
  RPath                    iClipRegion; // current clip path as region
  
  //
  // graphics state
  //
  CGraphicsContext* iGraphicsGc;
  TInt              iFlatness;
  TRgb              iFillColor;
  TRgb              iStrokeColor;
  TSize             iLineWidth;
  CGraphicsContext::TPenStyle iLineDash;
 
  //
  // text state
  //
  CGraphicsContext* iTextGc;
  CGDITextView*     iTextView;
  CGDIFontCache*    iFonts;
};
#endif
