// CEdgeTable.cpp
//
// Original code by Michael Abrash, Dr. Dobbs Journal #177, June 1991
//
// Ported to EPOC by Sander van der Wal, 12 june 2000
// Adapted to add rectangles to a region.
// Adapted to do winding fill rule as well
//
// $Id: edgetable.cpp 1.1 2000-09-17 13:20:24+02 svdwal Exp svdwal $
//
// $Log: edgetable.cpp $
// Revision 1.1  2000-09-17 13:20:24+02  svdwal
// Initial checkin
//

#include "EdgeTable.h"

///////////////////////////////////
//
// CEdgeState
//
///////////////////////////////////
struct TEdgeState
{
  TEdgeState* iNextEdge;

  TInt iX;
  TInt iStartY;
  TInt iWholePixelMove;
  TInt iXDirection;
  TInt iYDirection;
  TInt iErrorTerm;
  TInt iErrorTermAdjUp;
  TInt iErrorTermAdjDown;
  TInt iCount;
};


///////////////////////////////////
//
// CEdgeTable 
//
///////////////////////////////////
CEdgeTable::CEdgeTable() {}


void CEdgeTable::ConstructL(const CArrayFixFlat<TPoint>& aPolygon)
{
  TInt numHori = 0;

  iTable = (TEdgeState*) User::AllocL(aPolygon.Count() * sizeof(TEdgeState));

  TEdgeState* nextFreeEdgeStruct = iTable;
  
  for (TInt i = 0; i < aPolygon.Count(); i++) {

    // the edge runs from the current point to the previous one
    TPoint start(aPolygon.At(i));
    TPoint end(aPolygon.At( i == 0 ? aPolygon.Count()-1 : i-1));
    
    TInt yDirection;
    // make sure the edge runs from top to bottom
    if (start.iY > end.iY) {
      TPoint temp(start);
      start = end;
      end = temp;
      yDirection = -1;
    }
    else
      yDirection = 1;
   
    // skip if this can't ever be an active edge (has 0 height)
    TInt deltaY = end.iY - start.iY;
    if (deltaY != 0) {
      
      // allocate space for the edge
      TEdgeState* newEdgePtr = nextFreeEdgeStruct++;
      
      // direction in which Y moves
      newEdgePtr->iYDirection = yDirection;

      // direction in which X moves
      TInt deltaX = end.iX - start.iX;
      newEdgePtr->iXDirection = deltaX > 0 ? 1 : -1;

      TInt width = Abs(deltaX);
      newEdgePtr->iX = start.iX;
      newEdgePtr->iStartY = start.iY;
      newEdgePtr->iCount = deltaY;
      newEdgePtr->iErrorTermAdjDown = deltaY;
      if (deltaX >= 0) 
        // initial error term going L->R
        newEdgePtr->iErrorTerm = 0;
      else
        // initial error term going R->L
        newEdgePtr->iErrorTerm = -deltaY+1;
      if (deltaY >= width) {
        // Y-Major edge
        newEdgePtr->iWholePixelMove = 0;
        newEdgePtr->iErrorTermAdjUp = width;
      }
      else {
        // X-Major edge
        newEdgePtr->iWholePixelMove = (width / deltaY) * newEdgePtr->iXDirection;
        newEdgePtr->iErrorTermAdjUp = width % deltaY;
      }

      // link new edge into the GET so that the edge list is
      // still sorted by Y coordinate and the by X coordinate for all
      // edges with the same Y coordinate
      TEdgeState** followingEdgeLink = &iGETPtr;
      for (;;) {
        TEdgeState* followingEdge = *followingEdgeLink;
        if ((followingEdge == 0) ||
            (followingEdge->iStartY > start.iY) ||
            (followingEdge->iStartY == start.iY && followingEdge->iX >= start.iX)) {
          newEdgePtr->iNextEdge = followingEdge;
          *followingEdgeLink = newEdgePtr;
          break;
        }
        followingEdgeLink = &followingEdge->iNextEdge;
      }

    } // if active edge
    else {
      numHori++;
    }

  } // for all segments

  // initialioze the active edge table to empty
  iAETPtr = 0;
}


CEdgeTable::~CEdgeTable()
{ 
  User::Free(iTable);
}


TInt CEdgeTable::StartY() const { return iGETPtr->iStartY; }


void CEdgeTable::AddScanlineToRegion(TInt aYToScan, RRegion& aRegion, CGraphicsContext::TFillRule aFillRule)
{
  MoveXSortedToAET(aYToScan);
  RectOutAET(aYToScan, aRegion, aFillRule);
}


void CEdgeTable::NextScanline()
{
  AdvanceAET();
  XSortAET();
}


void CEdgeTable::XSortAET()
{
  if (iAETPtr != 0) {
    TBool swapOccured;
    do {
      swapOccured = EFalse;
      TEdgeState* currentEdge;
      TEdgeState** currentEdgePtr = &iAETPtr;
      while ((currentEdge= *currentEdgePtr)->iNextEdge != 0) {
        if (currentEdge->iX > currentEdge->iNextEdge->iX) {
          // the second edge has a lower X than the first;
          // swap them in the AET
          TEdgeState* tempEdge = currentEdge->iNextEdge->iNextEdge;
          *currentEdgePtr= currentEdge->iNextEdge;
          currentEdge->iNextEdge->iNextEdge = currentEdge;
          currentEdge->iNextEdge = tempEdge;
          swapOccured = ETrue;
        }
        currentEdgePtr = &(*currentEdgePtr)->iNextEdge;
      } // while
    } while (swapOccured);
  }
}


void CEdgeTable::AdvanceAET()
{
  TEdgeState* currentEdge;
  TEdgeState** currentEdgePtr = &iAETPtr;
  while ((currentEdge = *currentEdgePtr) != 0) {
    // count off one scan line for this edge
    if ((--(currentEdge->iCount)) == 0) {
      // this edge is finished, so remove it from the AET
      *currentEdgePtr = currentEdge->iNextEdge;
    }
    else {
      // advance the edge's X coordinate by minimum move
      currentEdge->iX += currentEdge->iWholePixelMove;
      // determine whether it's time for X to advange one extra
      if ((currentEdge->iErrorTerm += currentEdge->iErrorTermAdjUp) > 0) {
        currentEdge->iX += currentEdge->iXDirection;
        currentEdge->iErrorTerm -= currentEdge->iErrorTermAdjDown;
      }
      currentEdgePtr = &currentEdge->iNextEdge;
    }
  } // while
}


void CEdgeTable::MoveXSortedToAET(TInt aYToMove)
{
  // the GET is Y sorted. Any edges that start at the desired Y
  // coordinate will be first in the GET, so we'll move edges from
  // the GET to AET until the first edge left in the GET is no longer
  // at the desired Y coordinate. Also, the GET is X sorted within
  // each Y coordinate, so each successive edge we add to the AET is
  // guaranteed to belong later in the AET than the one just added
  
  TEdgeState* AETEdge;
  TEdgeState** AETEdgePtr = &iAETPtr;
  while ((iGETPtr != 0) && (iGETPtr->iStartY == aYToMove)) {
    
    TInt currentX = iGETPtr->iX;
    
    // link the new edge into the AET so that the AET is still
    // sorted by X coordinate

    for (;;) {
      AETEdge = *AETEdgePtr;
      if ((AETEdge == 0) || (AETEdge->iX  >= currentX)) {
        TEdgeState* tempEdge = iGETPtr->iNextEdge;
        *AETEdgePtr = iGETPtr; // link the edge into the AET
        iGETPtr->iNextEdge = AETEdge;
        AETEdgePtr = &iGETPtr->iNextEdge;
        iGETPtr = tempEdge; // unlink the edge from the GET
        break;
      }
      else {
        AETEdgePtr = &AETEdge->iNextEdge;
      }
    }
  }
}


void CEdgeTable::RectOutAET(TInt aYToScan, RRegion& aRegion, CGraphicsContext::TFillRule aFillRule)
{
  // scan through the AET, drawing line segments as each pair if edge
  // crossings is encountered. The nearest pixel on or to the right
  // of left edges is drawn, and the nearest pixel to the left of but 
  // not on right edges is drawn

  register TEdgeState* currentEdge = iAETPtr;
  if (aFillRule == CGraphicsContext::EAlternate) {
    while (currentEdge != 0) {
      TInt leftX = currentEdge->iX;
      currentEdge = currentEdge->iNextEdge;
      aRegion.AddRect(TRect(leftX, aYToScan, currentEdge->iX, aYToScan+1));
      currentEdge = currentEdge->iNextEdge;
    }
  }
  else { // CGraphicsContext::EWinding
    register TInt winding = 0;
    while (currentEdge != 0) {
        TInt leftX = currentEdge->iX;
        winding += currentEdge->iYDirection;
        currentEdge = currentEdge->iNextEdge;
        if (winding != 0)
          aRegion.AddRect(TRect(leftX, aYToScan, currentEdge->iX, aYToScan+1));
    }
  }
  aRegion.Tidy();
}

