// GDIOutputDev.cpp
//
// Copyright 1999-2000  Sander van der Wal
//
// $Id: gdioutputdev.cpp 1.5 2000-09-27 09:37:38+02 svdwal Exp svdwal $
//
// $Log: gdioutputdev.cpp $
// Revision 1.5  2000-09-27 09:37:38+02  svdwal
// DrawChar() optimised a bit
//
// Revision 1.4  2000-09-25 21:00:19+02  svdwal
// Use SetClippingRect() in updateCropBox()
//
// Revision 1.3  2000-09-24 21:16:11+02  svdwal
// Bug fixes related to clipping region
//
// Revision 1.2  2000-09-21 13:56:47+02  svdwal
// Some bug fixes
//
// Revision 1.1  2000-09-17 13:20:27+02  svdwal
// Initial checkin
//

#ifdef __GNUC__
#pragma implementation
#endif

// Interface
#include "GDIOutputDev.h"

// BASE
#ifndef __E32MATH_H__
#include <e32math.h>
#endif
 
// GDI
#ifndef __FBS_H__
#include <fbs.h>
#endif

// CONE
#ifndef __COEMAIN_H__
#include <coemain.h>
#endif

// EIKON
#ifndef __EIKENV_H__
#include <eikenv.h>
#endif
#ifndef __EIKMSG_H__
#include <eikmsg.h>
#endif

// View
#include "GDIView.h"
#include "EdgeTable.h"

// PDF
#include "Error.h"
#include "GfxState.h"
#include "GfxFont.h"
#include "FontEncoding.h"

#include "PDFGLobal.h"
#include "Pdf.rsg"
#include "Pdf.hrh"

// PROFILING
#include "PROFILE.H"


// Local
static inline TInt round(double x) { return (TInt) (x+0.5); }

static const TInt KBusyInitialDelay = 100000; // 0.1 second



///////////////////////////////////
//
// RPath
//
///////////////////////////////////
struct TRun
{
  TRun(TInt aX, TInt aWidth): iX(aX), iWidth(aWidth), iNext(0) {}
  TInt iX;
  TInt iWidth;
  TRun* iNext;
};

void RPath::Set(const TRect& aRect)
{
  Clear();
  AddRect(aRect);
}


void RPath::AddPolygonL(const CPolygon& aPolygon, CGraphicsContext::TFillRule aFillRule)
{
  // often just a rectangle is passed
  if (aPolygon.IsRect()) {
    TRect rect(aPolygon.BoundingRect());
    AddRect(rect);
    Tidy();
    return;
  }
  // build the region:
  // iterate over the scanlines
  // add the scanline rectangles to the region

  // table of edges
  CEdgeTable* edges = new(ELeave) CEdgeTable();
  CleanupStack::PushL(edges);
  edges->ConstructL(aPolygon);

  // loop along until finished
  TInt y = edges->StartY();
  while (!edges->Empty()) {

    edges->AddScanlineToRegion(y, *this, aFillRule);
    edges->NextScanline();
    
    y++;

  } // for all scanlines

  CleanupStack::PopAndDestroy(); // edges

}


TBool RPath::Intersects(const TRect& aRect)
{
  if (IsEmpty())
    return EFalse;

  if (aRect.IsEmpty())
    return EFalse;

  for (TInt i = 0; i<Count(); i++) {
    if (aRect.Intersects(this->operator[](i)))
      return ETrue;
  }

  return EFalse;
}


void RPath::Intersect(const TRect& aRect)
{
  // create a region for the rect, then Intersect the region
  RRegion rect(aRect, 1);
  CleanupClosePushL(rect);
  RRegion::Intersect(rect);
  rect.Clear();
  CleanupStack::PopAndDestroy(); // recr
}

///////////////////////////////////
//
// TGDIOutputState
//
///////////////////////////////////
TGDIOutputState::TGDIOutputState(CGDIOutputDev* aDev)
: iNext(aDev->iSave),
  iClipPath  (aDev->iClipPath),
  iFillRule  (aDev->iFillRule),
  iClipRegion(aDev->iClipRegion),
  
  iGraphicsGc (aDev->iGraphicsGc),
  iFlatness   (aDev->iFlatness),
  iFillColor  (aDev->iFillColor),
  iStrokeColor(aDev->iStrokeColor),
  iLineWidth  (aDev->iLineWidth),
  iLineDash   (aDev->iLineDash),
  
  iTextGc  (aDev->iTextGc),
  iTextView(aDev->iTextView)
{
  aDev->iSave = this;
}


void TGDIOutputState::RestoreAndPop(CGDIOutputDev* aDev)
{
  // kill current state
  delete aDev->iTextGc;
  delete aDev->iTextView;
  delete aDev->iGraphicsGc;
  if (aDev->iClipPath != iClipPath) {
    aDev->iClipPath->ResetAndDestroy();
    aDev->iClipPath->Reset();
    delete aDev->iClipPath;
  }
  aDev->iClipRegion.Clear();
   
  // restore state
  aDev->iGraphicsGc = iGraphicsGc;
  aDev->iFlatness   = iFlatness;
  aDev->iFillColor  = iFillColor;
  aDev->iStrokeColor= iStrokeColor;
  aDev->iLineWidth  = iLineWidth;

  aDev->iTextGc   = iTextGc;
  aDev->iTextView = iTextView;

  aDev->iClipPath   = iClipPath;
  aDev->iFillRule   = iFillRule;
  aDev->iClipRegion = iClipRegion;

  // pop the stack
  TGDIOutputState* t = aDev->iSave;
  aDev->iSave = this->iNext;
  delete t;
}


///////////////////////////////////
//
// CGDIOutputDev
//
///////////////////////////////////
CGDIOutputDev::CGDIOutputDev(){}

CGDIOutputDev::CGDIOutputDev(TBool aRenderText, TBool aRenderLineDrawings, TBool aRenderImages)
  : iRenderText(aRenderText), 
    iRenderLineDrawings(aRenderLineDrawings),
    iRenderImages(aRenderImages)
{
  // Set the default substitution typeface names
  iFontNameSubst[0].iName.Copy(_L("Arial"));
  iFontNameSubst[1].iName.Copy(_L("Times New Roman"));
  iFontNameSubst[2].iName.Copy(_L("Courier New"));
  iFontNameSubst[3].iName.Copy(_L("Symbol Medium"));
}


CGDIOutputDev::~CGDIOutputDev()
{  
  endPage();
}


//
// OutputDev reimplementation
//

//----- initialization and control
void CGDIOutputDev::InitPageL()
{
  iFillColor = KRgbBlack;
  iStrokeColor = KRgbBlack;

  if (iRenderText) {   
    
    // create the font cache
    iFonts = new(ELeave) CGDIFontCache(iDev, iFontNameSubst);
    iFonts->ConstructL();

    iTextView = new(ELeave) CGDITextView();
  }

  if (iRenderLineDrawings || iRenderImages) {
    iGraphicsGc->SetBrushStyle(CGraphicsContext::ESolidBrush);
    iGraphicsGc->SetBrushColor(iFillColor);
    iGraphicsGc->SetPenColor(iStrokeColor);  
  }
}


void CGDIOutputDev::CheckTypefacesL()
{ 
  // Now look if there are better typefaces available
  TTypefaceSupport* ts = new(ELeave) TTypefaceSupport;
  for (TInt i = 0; i < iDev->NumTypefaces(); i++) {
    iDev->TypefaceSupport(*ts, i);
    if (ts->iTypeface.IsSymbol()) {
      /* do nothing */;
    }
    else if (!ts->iTypeface.IsProportional()) {
      // Courier replacement
      iFontNameSubst[2].iName = ts->iTypeface.iName;
    }
    else if (ts->iTypeface.IsSerif()) {
      // Times New Roman replacement
      iFontNameSubst[1].iName = ts->iTypeface.iName;
      break;
    }
    else {
      // Helvetica replacement
      iFontNameSubst[0].iName = ts->iTypeface.iName;
    }
  }
  delete ts;
}


// End a page. Gets called during cleanup too
void CGDIOutputDev::endPage()
{ 
  ClearPath();
  
  // clear state stack
  while (iSave)
    restoreState(0);
  
  delete iGraphicsGc;
  iGraphicsGc = 0;

  delete iTextGc;
  iTextGc = 0;
  
  if (iClipPath) {
    iClipPath->ResetAndDestroy();
    iClipPath->Reset();
    delete iClipPath;
  }
  iClipRegion.Clear();
  
  delete iTextView;
  iTextView = 0;

  delete iFonts;
  iFonts = 0;
}


//----- update graphics state
void CGDIOutputDev::updateAll(GfxState *state)
{
  updateFlatness(state);
  updateLineDash(state);
  updateLineWidth(state);
  updateFillColor(state);
  updateStrokeColor(state);
  updateFont(state);
  updateTextMat(state);
  updateRender(state);
}


void CGDIOutputDev::updateCropBox(GfxState* state, 
                                  double cropX1, double cropY1, double cropX2, double cropY2)
{
  
  double x,y;
  TRect cropBox;

  state->transform(cropX1, cropY1, &x, &y);
  cropBox.iTl.iX = round(x);
  cropBox.iTl.iY = round(y);
       
  state->transform(cropX2, cropY2, &x, &y);
  cropBox.iBr.iX = round(x);
  cropBox.iBr.iY = round(y);
  
  cropBox.Normalize();
  iClipRegion.Intersect(cropBox);
  if (iClipRegion.CheckError())
    // because we didn't use ForceError() the error must be a memory shortage
    User::LeaveNoMemory();

  if (iRenderText)
  //  SetTextClipPath();
    iTextGc->SetClippingRect(cropBox);

  if (iRenderLineDrawings || iRenderImages)
  //  SetGraphicsClipPath();
    iGraphicsGc->SetClippingRect(cropBox);

}


void CGDIOutputDev::updateCTM(GfxState *state, 
                              double /*m11*/, double /*m12*/,
                              double /*m21*/, double /*m22*/, 
                              double /*m31*/, double /*m32*/)
{
  updateLineDash(state);
  updateLineWidth(state);
  updateTextMat(state);
}


void CGDIOutputDev::updateLineDash(GfxState *state)
{
  if (!iRenderLineDrawings && !iRenderImages)
    return;

  double* dash;
  int     length;
  double  start;
  state->getLineDash(&dash, &length, &start);
  
  // not a very accurate translation
  if (dash == 0 || length<=1)
    iLineDash = CGraphicsContext::ESolidPen;
  else if (length == 2)
    iLineDash = CGraphicsContext::EDottedPen;
  else if (length == 4)
    iLineDash = CGraphicsContext::EDashedPen;
  else 
    iLineDash = CGraphicsContext::EDotDashPen;
  iGraphicsGc->SetPenStyle(iLineDash);
}


void CGDIOutputDev::updateFlatness(GfxState *state)
{
  if (!iRenderLineDrawings && !iRenderImages)
    return;

  if (state->getFlatness() <1)
    iFlatness = 1;
  else
    iFlatness = state->getFlatness() * state->getFlatness();
}


void CGDIOutputDev::updateLineWidth(GfxState *state)
{
  if (!iRenderLineDrawings && !iRenderImages)
    return;

  TInt lineWidth = round(state->getLineWidth());
  if (lineWidth == 0)
    // device-dependent line width is one pixel
    lineWidth = 1;
  
  iLineWidth = TSize(lineWidth, lineWidth);
  iGraphicsGc->SetPenSize(iLineWidth);
}


void CGDIOutputDev::updateFillColor(GfxState *state)
{
  GfxColor* c = state->getFillColor();
#ifdef FIXCOLOR
  iFillColor = TRgb(c->getR(), c->getG(), c->getB());
#else
  iFillColor = TRgb(TInt(c->getR()*255), TInt(c->getG()*255), TInt(c->getB()*255));
#endif

  if (iRenderLineDrawings || iRenderImages)
    iGraphicsGc->SetBrushColor(iFillColor);
  
  if (iRenderText)
    iTextView->UpdateFillColor(iFillColor, iTextGc);
}


void CGDIOutputDev::updateStrokeColor(GfxState *state)
{
  GfxColor* c = state->getStrokeColor();
#ifdef FIXCOLOR
  iStrokeColor = TRgb(c->getR(), c->getG(), c->getB());
#else
  iStrokeColor = TRgb(TInt(c->getR()*255), TInt(c->getG()*255), TInt(c->getB()*255));
#endif

  if (iRenderLineDrawings || iRenderImages)
    iGraphicsGc->SetPenColor(iStrokeColor);
  
  if (iRenderText)
    iTextView->UpdateStrokeColor(iStrokeColor, iTextGc);
}

//----- update text state
void CGDIOutputDev::updateFont(GfxState *state)
{
  PROFILE_START(CGDIOUTPUTDEV__UPDATEFONT);

  if (iRenderText) {
    if (state->getFont())
      iTextView->UpdateFont(state, iTextGc, iFonts);
  }
  
  PROFILE_STOP(CGDIOUTPUTDEV__UPDATEFONT);
}


void CGDIOutputDev::updateTextMat(GfxState *state)
{
  if (!iRenderText)
    return;
  
  double dx, dy;
  double xt, yt;
  // find out if the text is rotated
  state->textTransformDelta(1, 0, &dx, &dy);
  state->transformDelta(dx, dy, &xt, &yt);
  TInt newRotate;
  if (xt > 0 && xt > yt)
    newRotate = 0;
  else if (yt > 0 && yt > xt)
    newRotate = 90;
  else if (xt < 0 && xt < yt)
    newRotate = 180;
  else
    newRotate = 270;
 
  if (newRotate != iTextView->Rotate())
    CreateTextViewL(newRotate);
}

void CGDIOutputDev::updateRender(GfxState* state)
{
  if (!iRenderText)
    return;

  iTextView->UpdateRender(state->getRender(), iTextGc);
}


void CGDIOutputDev::fill(GfxState *state, GBool strokeToo)
{
  ConvertPathL(state);
  
  if (!iRenderLineDrawings)
    return;
 
  // By setting the pen color to the fill color as well, an extra line is drawn around
  // the polygon, effectively making polygons one pixel wide visible
  iGraphicsGc->SetPenColor(strokeToo ? iStrokeColor:iFillColor);

  if (iClipPath)
    MakeClipRegion();
  
  FillPathL(CGraphicsContext::EWinding);
   
  iGraphicsGc->SetPenColor(iStrokeColor);
}


void CGDIOutputDev::eoFill(GfxState *state, GBool strokeToo)
{
  ConvertPathL(state);
  
  if (!iRenderLineDrawings)
    return;
  
  // By setting the pen color to the fill color as well, an extra line is drawn around
  // the polygon, effectively making polygons one pixel wide visible
  iGraphicsGc->SetPenColor(strokeToo ? iStrokeColor:iFillColor);
 
  if (iClipPath)
    MakeClipRegion();
  
  FillPathL(CGraphicsContext::EAlternate);
  
  // reset pen color setting
  iGraphicsGc->SetPenColor(iStrokeColor);
}


void CGDIOutputDev::ConvertPathL(GfxState* aState)
{
  // start building the path, cache it for stroking and/or clipping
  ClearPath();
  iGfxPath = aState->getPath();
  TInt nSubpaths = iGfxPath->getNumSubpaths();
  iPath = new(ELeave) CArrayPtrFlat<CPolygon>(nSubpaths);
  iPath->SetReserveL(nSubpaths); // so InsertL() does not leave

  // Get subpaths.  
  for (register TInt i = 0; i<nSubpaths; i++)
    iPath->InsertL(i, GetSubpathL(iGfxPath->getSubpath(i), aState));
}


CPolygon* CGDIOutputDev::GetSubpathL(GfxSubpath *aSubpath, GfxState* aState)
{
  // draw bezier curves or straight lines 
  TInt nPoints = aSubpath->getNumPoints();
  CPolygon* polygon = new(ELeave) CPolygon(nPoints);
  CleanupStack::PushL(polygon);
  
  TPoint p0;
  register TInt j = 0;
  while (j < nPoints) {
    double x, y;
    if (j >= 1 && aSubpath->getCurve(j)) {

      // add bezier curve
      
      // first point of curve is last point of previous segment
      // aState->transform(aSubpath->getX(j-1),   aSubpath->getY(j-1), &x, &y);
      // TPoint p0(round(x), round(y));
        
      aState->transform(aSubpath->getX(j),   aSubpath->getY(j), &x, &y);
      TPoint p1(round(x), round(y));
       
      aState->transform(aSubpath->getX(j+1), aSubpath->getY(j+1), &x, &y);
      TPoint p2(round(x), round(y));
     
      aState->transform(aSubpath->getX(j+2), aSubpath->getY(j+2), &x, &y);
      TPoint p3(round(x), round(y));

      polygon->AddBezierL(p0, p1, p2, p3, iFlatness);
      p0 = p3;
      j += 3;
    } 
    else {
      aState->transform(aSubpath->getX(j), aSubpath->getY(j), &x, &y);
      p0.iX = round(x);
      p0.iY = round(y);
      polygon->AddLineL(p0);
      j++;
    }
  } // while
  CleanupStack::Pop(); // subpath
  return polygon;
}


//----- path clipping
void CGDIOutputDev::clip(GfxState* state)
{
  if (!iClip)
    return;
  DoClipL(state, CGraphicsContext::EWinding);
}


void CGDIOutputDev::eoClip(GfxState * state)
{
  if (!iClip)
    return;
  DoClipL(state, CGraphicsContext::EAlternate);
}


void CGDIOutputDev::DoClipL(GfxState* aState, CGraphicsContext::TFillRule aFillRule)
{
  // if the path hasn't been filled or stroked, build it first
  if (iGfxPath != aState->getPath())
    ConvertPathL(aState);
  
  //
  // delay the building of a region as long as possible
  //

  // if the path is just a rectangle, add rectangle to the region
  if ((iPath->Count() == 1) && iPath->At(0)->IsRect()) {

    iClipRegion.Intersect(iPath->At(0)->BoundingRect());
    if (iClipRegion.CheckError())
      // because we didn't use ForceError() the error must be a memory shortage
      User::LeaveNoMemory();
    ClearPath();
    return;
  
  }

  // if the current polygonpath is still there, add it to the current region first
  if (iClipPath)
    MakeClipRegion();

  __ASSERT_DEBUG(iClipPath == 0, User::Invariant());

  // set the clipping path as the path
  iClipPath = iPath;
  iFillRule = aFillRule;
  
  // clear the path without destroying it
  iPath = 0;
  iGfxPath = 0;
}


void CGDIOutputDev::MakeClipRegion()
{
  __ASSERT_DEBUG(iClip, User::Invariant());

  // create a region for this path
  RPath clipPath;
  CleanupClosePushL(clipPath);
  
  // Add polygons to clippath
  for (TInt i = 0; i<iClipPath->Count(); i++) {
    CPolygon* polygon = iClipPath->At(i);
    clipPath.AddPolygonL(*polygon, iFillRule);
  }
  
  CleanupStack::Pop(); // clipPath
 
  // if the saved clippath is identical to this one, do not destroy it
  // because it is shared between (saved) states
  if (iSave->iClipPath != iClipPath) {
    iClipPath->ResetAndDestroy();
    iClipPath->Reset();
    delete iClipPath;
  }
  iClipPath = 0;

  // intersect the two regions
  //if (iClipState != EClipRegion)
  //  iClipRegion.Set(iClipRect);
  iClipRegion.Intersect(clipPath);
  if (iClipRegion.CheckError())
    // because we didn't use ForceError() the error must be a memory shortage
    User::LeaveNoMemory();
  clipPath.Clear();
}


void CGDIOutputDev::SetTextClipPath() {}


void CGDIOutputDev::SetGraphicsClipPath() {}


void CGDIOutputDev::ClearPath()
{
  if (iPath) {
    iPath->ResetAndDestroy();
    iPath->Reset();
    delete iPath;
    iPath = 0;
  }
  iGfxPath = 0;
}


//----- text drawing
void CGDIOutputDev::beginString(GfxState*state, GString* /* s */)
{
  if (!iRenderText)
    return;
  
  if (iClipPath)
    MakeClipRegion();
  
  iTextView->BeginString(state);
}


void CGDIOutputDev::endString(GfxState* state)
{
  if (!iRenderText)
    return;

  //check for invisible text -- this is used by Acrobat Capture
  if ((state->getRender() & 3) == 3)
    return;

  iTextView->EndString(state);
}


void CGDIOutputDev::drawChar(GfxState * state,
                             double x, double y,
                             double dx, double /* dy */, 
                             Guchar c)
{
  PROFILE_START(CGDIOUTPUTDEV__DRAWCHAR);

  if (!iRenderText)
    return;

  //check for invisible text -- this is used by Acrobat Capture
  if ((state->getRender() & 3) == 3)
    return;
  
  // calculate position
  double x1, y1;
  state->transform(x, y, &x1, &y1);
  iTextView->iPos.iX = round(x1);
  iTextView->iPos.iY = round(y1);
  
  iTextView->DrawChar(iTextGc, c, TInt(dx));

  PROFILE_STOP(CGDIOUTPUTDEV__DRAWCHAR);
}



void CGDIOutputDev::drawString(GfxState *state, GString *s)
{
  PROFILE_START(CGDIOUTPUTDEV__DRAWSTRING);

  if (!iRenderText)
    return;

  // check for invisible text -- this is used by Acrobat Capture
  if ((state->getRender() & 3) == 3)
    return;

  if (iClipPath)
    MakeClipRegion();

  iTextView->SetPos(state);
  iTextView->DrawString(state, iTextGc, s);

  PROFILE_STOP(CGDIOUTPUTDEV__DRAWSTRING);
}


void CGDIOutputDev::GetImagePlacementInfo(GfxState *aState, TRect& aRect,
                                          TBool& aRotate, TBool& aXFlip, TBool& aYFlip)
{
  // get image position and size in device coordinates
  double xt, yt, wt, ht;
  aState->transform(0, 0, &xt, &yt);
  aState->transformDelta(1, 1, &wt, &ht);
  if (wt > 0) {
    aRect.iTl.iX = round(xt);
    aRect.SetWidth(round(wt));
  } else {
    aRect.iTl.iX = round(xt + wt);
    aRect.SetWidth(round(-wt));
  } 
  if (ht > 0) {
    aRect.iTl.iY = round(yt);
    aRect.SetHeight(round(ht));
  } else {
    aRect.iTl.iY = round(yt + ht);
    aRect.SetHeight(round(-ht));
  }
  aRect.Normalize();

  // find out whether and how the image is flipped and/or rotated
  aState->transformDelta(1, 0, &xt, &yt);
  aRotate = Abs(xt) < Abs(yt);
  if (aRotate) {
    aXFlip = ht < 0;
    aYFlip = wt > 0;
  } else {
    aXFlip = wt < 0;
    aYFlip = ht > 0;
  } 
}


void CGDIOutputDev::DoDrawImageL(Stream *aStr, TSize aImgSize, GfxImageColorMap *aColorMap,
                                 TBool aRotate, TBool aXFlip, TBool aYFlip,
                                 TBool aDitherImages,
                                 const TRect& aDevRect,
                                 CWsBitmap* aDstBitmap, TDisplayMode aDisplayMode,
                                 const TRect& aClipRect)
{
  // setup
  TBool dither = EFalse;
  if (aDitherImages ) {
    TInt comps = aColorMap->getNumPixelComps();
    TInt bits = aColorMap->getBits(); 
    switch (aDisplayMode) {
    case EGray2:    dither = (comps >  1) || (bits > 1); break;
    case EGray4:    dither = (comps >  1) || (bits > 2); break;
    case EGray16:   dither = (comps >  1) || (bits > 4); break;
    case EGray256:  dither = (comps >  1) || (bits > 8); break;
    case EColor16:  dither = (comps >= 3) || (bits > 2); break;
    case EColor256: dither = (comps >= 3) || (bits > 4); break;
    case EColor4K:  dither = (comps >= 3) || (bits > 4); break;
    case EColor64K: dither = (comps >= 3) || (bits > 4); break;
    case EColor16M: 
    case ERgb:      dither = (comps >= 3) || (bits > 8); break;
    default:
      break;
    }
  } 

  // initialize the image stream
  ImageStream* imgStr = new(ELeave) ImageStream(aStr, aImgSize.iWidth, aColorMap->getNumPixelComps(), aColorMap->getBits());
  CleanupStack::PushL(imgStr);
  imgStr->ConstructL();
  imgStr->reset();

  // the size of the unrotated image in device measurements
  TSize devSize(aRotate ? aDevRect.Height(): aDevRect.Width(),
                aRotate ? aDevRect.Width() : aDevRect.Height());

  // Bresenham parameters
  TInt px1 = devSize.iWidth / aImgSize.iWidth;
  TInt px2 = devSize.iWidth - px1 * aImgSize.iWidth;
  TInt py1 = devSize.iHeight / aImgSize.iHeight;
  TInt py2 = devSize.iHeight - py1 * aImgSize.iHeight;

  // allocate error diffusion accumulators
  TGDIRgb err;
  TGDIRgb* errDown = 0;
  TGDIRgb* errRight = 0;
  if (dither) {
    errDown  = (TGDIRgb*) User::AllocL(devSize.iWidth * sizeof(TGDIRgb));
    CleanupStack::PushL(errDown);
    errRight= (TGDIRgb*) User::AllocL((py1+1) * sizeof(TGDIRgb));
    CleanupStack::PushL(errRight);
    for (register TInt j = 0; j < devSize.iWidth; j++)
      errDown[j] = TGDIRgb();
  }
  
  GfxColor gfxColor;

  if (aRotate) {
    // pointer to member function IncXPos, DecXPos, IncYPos and DecYPos
    typedef void (TBitmapUtil::*TBitmapUtilNextPos)();
    TBitmapUtilNextPos nextXPos = aXFlip ? &TBitmapUtil::DecYPos : &TBitmapUtil::IncYPos;
  
    // first line (column)
    TBitmapUtil* util = new(ELeave) TBitmapUtil(aDstBitmap);
    TPoint pos(aDevRect.iTl.iX + (aYFlip ? devSize.iHeight-1: 0), 
               aDevRect.iTl.iY + (aXFlip ? devSize.iWidth -1: 0));
    util->Begin(pos);

    TInt qy = 0;
    for (TInt i=0; i < aImgSize.iHeight; i++) {

      // vertical (horizontal) Bresenham
      TInt dy = py1;
      if ((qy += py2) >= aImgSize.iHeight) {
        ++dy;
        qy -= aImgSize.iHeight;
      }
     
      TBool clip = (dy == 0);
      if (!clip) {
        // first column (line)
        clip = pos.iX < aClipRect.iTl.iX || aClipRect.iBr.iX <= pos.iX;
      }

      if (clip) {
        // drop a line (column)  
        imgStr->skipLine();
      } 
      else {
        // draw a line (column)

        TBitmapUtilNextPos nextYPos= aYFlip ? &TBitmapUtil::DecXPos : &TBitmapUtil::IncXPos;
        TBitmapUtilNextPos prevYPos= aYFlip ? &TBitmapUtil::IncXPos : &TBitmapUtil::DecXPos; 
     
        TInt x = aXFlip ? devSize.iWidth - 1 : 0;
        pos.iY = aDevRect.iTl.iY + x;
        util->SetPos(pos);

        // clear error accumulator
        if (dither) {
          for (register TInt j = 0; j <= py1; j++)
            errRight[j] = TGDIRgb();
        }

        // read a line (column)
        TInt qx = 0;
        for (register TInt j= 0; j< aImgSize.iWidth; j++) {    

          // horizontal (vertical) Bresenham
          TInt dx = px1;
          if ((qx += px2) >= aImgSize.iWidth) {
            ++dx;
            qx -= aImgSize.iWidth;
          }

          Guchar pixBuf[4];
          imgStr->getPixel(pixBuf);
          if (dx <= 0)
            continue;

          if (aDevRect.iTl.iY + x < aClipRect.iTl.iY || aClipRect.iBr.iY <= aDevRect.iTl.iY + x) {
            for (register TInt ix = 0; ix < dx; ix++)
              (util->*nextXPos)();
          }
          else {
            // draw image pixel
            register TInt pixel;
            aColorMap->getColor(pixBuf, &gfxColor);
            TGDIRgb rgbColor(gfxColor.getR(), gfxColor.getG(), gfxColor.getB());
            if (dither) { 
              rgbColor = rgbColor + errRight[0] + errDown[x];
              pixel = FindPixel(rgbColor, aDisplayMode, err);
              errRight[0] = errDown[x] = TGDIRgb(err.Red()/2, err.Green()/2, err.Blue()/2);
            }
            else 
              pixel = CGDIView::FindPixel(rgbColor.Rgb(), aDisplayMode);
        
            if (dy == 1) {
              for (register TInt ix = 0; ix < dx; ix++) {
                util->SetPixel(pixel);
                (util->*nextXPos)();
              }
            }
            else {
              for (register TInt ix = 0; ix < dx; ix++) {
                for (register TInt iy = 0; iy < dy; iy++) {    
                  util->SetPixel(pixel);
                  (util->*nextYPos)();
                }
                util->SetPixel(pixel);
                (util->*nextXPos)();
  
                // swap direction of y
                TBitmapUtilNextPos tmp = nextYPos;
                nextYPos = prevYPos;
                prevYPos = tmp;
              }
            }
          }
          x += aXFlip ? -dx:dx;

        } // for (j=0; ...
      
      }// if (clip)
      // next column
      pos.iX += aYFlip ? -dy: dy;  
  
   } // for (i=0; ...
  
    util->End();
    delete util;
  
  } // if aRotate 
  else {
    // eventually this should be done using scanlines

    // pointer to member function IncXPos, DecXPos, IncYPos and DecYPos
    typedef void (TBitmapUtil::*TBitmapUtilNextPos)();
    TBitmapUtilNextPos nextXPos = aXFlip ? &TBitmapUtil::DecXPos : &TBitmapUtil::IncXPos;
   
    // first line (column)
    TBitmapUtil* util = new(ELeave) TBitmapUtil(aDstBitmap);   
    TPoint pos(aDevRect.iTl.iX + (aXFlip ? devSize.iWidth -1: 0), 
               aDevRect.iTl.iY + (aYFlip ? devSize.iHeight-1: 0));
    util->Begin(pos);
    
    TInt qy = 0;
    for (TInt i=0; i < aImgSize.iHeight; i++) {

      // vertical (horizontal) Bresenham
      TInt dy = py1;
      if ((qy += py2) >= aImgSize.iHeight) {
        ++dy;
        qy -= aImgSize.iHeight;
      }
     
      TBool clip = (dy == 0);
      if (!clip) {
        // first column (line)
        clip = pos.iY < aClipRect.iTl.iY || aClipRect.iBr.iY <= pos.iY;
      }

      if (clip) {
        // drop a line (column)  
        imgStr->skipLine();
      } 
      else {
        // draw a line (column)
        
        TBitmapUtilNextPos nextYPos = aYFlip ? &TBitmapUtil::DecYPos : &TBitmapUtil::IncYPos;
        TBitmapUtilNextPos prevYPos = aYFlip ? &TBitmapUtil::IncYPos : &TBitmapUtil::DecYPos;
     
        TInt x = aXFlip ? devSize.iWidth - 1 : 0;
        pos.iX = aDevRect.iTl.iX + x;
        util->SetPos(pos);

        // clear error accumulator
        if (dither) {
          for (register TInt j = 0; j <= py1; j++)
            errRight[j] = TGDIRgb();
        }
 
        TInt qx = 0;
        // read a line (column)
        for (register TInt j= 0; j< aImgSize.iWidth; j++) {    

          // horizontal (vertical) Bresenham
          TInt dx = px1;
          if ((qx += px2) >= aImgSize.iWidth) {
            ++dx;
            qx -= aImgSize.iWidth;
          }

          Guchar pixBuf[4];
          imgStr->getPixel(pixBuf);
          if (dx <= 0)
            continue;

          if (aDevRect.iTl.iX + x < aClipRect.iTl.iX || aClipRect.iBr.iX <= aDevRect.iTl.iX + x) {
            for (register TInt ix = 0; ix < dx; ix++)
              (util->*nextXPos)();
          }
          else {
            // draw image pixel
            register TInt pixel;
            aColorMap->getColor(pixBuf, &gfxColor);
            TGDIRgb rgbColor(gfxColor.getR(), gfxColor.getG(), gfxColor.getB());
            if (dither) { 
              rgbColor = rgbColor + errRight[0] + errDown[x];
              pixel = FindPixel(rgbColor, aDisplayMode, err);
              errRight[0] = errDown[x] = TGDIRgb(err.Red()/2, err.Green()/2, err.Blue()/2);
            }
            else 
              pixel = CGDIView::FindPixel(rgbColor.Rgb(), aDisplayMode);
        
            if (dy == 1) {
              for (register TInt ix = 0; ix < dx; ix++) {
                util->SetPixel(pixel);
                (util->*nextXPos)();
              }
            }
            else {
              for (register TInt ix = 0; ix < dx; ix++) {
                for (register TInt iy = 0; iy < dy; iy++) {    
                  util->SetPixel(pixel);
                  (util->*nextYPos)();
                }
                util->SetPixel(pixel);
                (util->*nextXPos)();
 
                // swap direction of y
                TBitmapUtilNextPos tmp = nextYPos;
                nextYPos = prevYPos;
                prevYPos = tmp;
              }
            }
          }
          x += aXFlip ? -dx:dx;

        } // for (j=0; ...
      } // if (clip)
      // next line (column)
      pos.iY += aYFlip ? -dy: dy;
  
    } // for (i=0; ...

    util->End();
    delete util;
  }
  
  if (dither)
    CleanupStack::PopAndDestroy(2); // errRight, errDown

  CleanupStack::PopAndDestroy(); //  imgStream
}


TInt CGDIOutputDev::FindPixel(const TGDIRgb& aColor, TDisplayMode aDisplayMode, TGDIRgb& aError) {
  
  TInt pixel = CGDIView::FindPixel(aColor.Rgb(), aDisplayMode);
 
  switch (aDisplayMode) {
  case EGray2: {
      // error is the difference between 256 graylevels and 2 gray levels
      TInt p256 = aColor.Rgb().Gray256();
      aError = TGDIRgb(p256 - 255*pixel, p256 - 255*pixel, p256 - 255*pixel);
      return pixel;
    }
    break;
  case EGray4: {
      // error is the difference between 256 graylevels and 4 gray levels
      TInt p256 = aColor.Rgb().Gray256();
      aError = TGDIRgb(p256 - 85*pixel, p256 - 85*pixel, p256 - 85*pixel);
      return pixel;
    }
    break;
  case EGray16: {
      // error is the difference between 256 graylevels and 16 gray levels
      TInt p256 = aColor.Rgb().Gray256();
      aError = TGDIRgb(p256 - 17*pixel, p256 - 17*pixel, p256 - 17*pixel);
      return pixel;
    }
    break;
  case EGray256:
    aError = TGDIRgb();
    return aColor.Rgb().Gray256();
    break;
  case EColor16: {
      TInt pixel = aColor.Rgb().Color16();
      TRgb back = TRgb::Color16(pixel);
      aError.SetRed  (aColor.Red()  - back.Red());
      aError.SetGreen(aColor.Green()- back.Green()); 
      aError.SetBlue (aColor.Blue() - back.Blue());
      return pixel;
    }
    break;
  case EColor256: {
      TRgb back = TRgb::Color256(pixel);
      aError.SetRed  (aColor.Red()  - back.Red()  );
      aError.SetGreen(aColor.Green()- back.Green()); 
      aError.SetBlue (aColor.Blue() - back.Blue() );
    return pixel;
    }
    break;
  case EColor64K:
    aError.SetRed  (aColor.Red()  &0x7);
    aError.SetGreen(aColor.Green()&0x3); 
    aError.SetBlue (aColor.Blue() &0x7);
    return pixel;
    break;
  case EColor16M: case ERgb:
    aError = TGDIRgb(0,0,0);
    return pixel;
    break;
  case EColor4K:
    aError.SetRed  (aColor.Red()  &0xf);
    aError.SetGreen(aColor.Green()&0xf); 
    aError.SetBlue (aColor.Blue() &0xf);
    return pixel;
    break;
  case ENone:
    break;
  }
  return 0;
}
