// GDIView.cpp
//
// Copyright 1999-2000  Sander van der Wal
//
// $Id: gdiview.cpp 1.3 2000-09-27 09:37:37+02 svdwal Exp svdwal $
//
// $Log: gdiview.cpp $
// Revision 1.3  2000-09-27 09:37:37+02  svdwal
// DrawChar() optimised a bit
//
// Revision 1.2  2000-09-24 21:15:24+02  svdwal
// - BoundingRect() fixed
// - IsRect() also true for four edges
//
// Revision 1.1  2000-09-17 13:20:23+02  svdwal
// Initial checkin
//

#ifdef __GNUC__
#pragma implementation
#endif

// Interface
#include "GDIView.h"

// BASE
#ifndef __E32MATH_H__
#include <e32math.h>
#endif
 
// CONE
#ifndef __COEMAIN_H__
#include <coemain.h>
#endif

// GooLib
#include "GString.h"

// PDF
#include "Error.h"
#include "GfxState.h"
#include "GfxFont.h"
#include "PDFGLobal.h"

// PROFILING
#include "PROFILE.H"

// Local
static inline TInt round(double x) { return (TInt) (x+0.5); }
static inline TBool isxdigit(TInt c) { return (('0'<=c && c<='9') || ('A'<=c && c<='F') || ('a'<=c && c<='f')); }
static inline TBool isdigit(TInt c) { return ('0'<=c && c<='9'); }
static inline TInt HexToInt(char c)
{
  if ('0'<=c && c<='9')
    return c - '0';
  else if ('A'<=c && c<='F')
    return c - 'A' + 10;
  else if ('a'<=c && c<='f')
    return c - 'a' + 10;
  else
    return -1;
}

static TInt HexToInt(char c1, char c2)
{
  return HexToInt(c1)*16+HexToInt(c2);  
}

static TInt atoi(register const char* s)
{
  TInt res = 0;
  register char c = *s;
  while (isdigit(c)) {
    res = res*10 + c - '0';
    c=*++s;
  }
  return res;  
}


////////////////////////////////////////////////
//
// TGfxFontSpec
//
// A GfxFont in Epoc terminology
//
////////////////////////////////////////////////
class TGfxFontSpec: public TFontSpec
{
public:
  TGfxFontSpec(): TFontSpec() {};
  TGfxFontSpec(const TDesC& aName, TFontStrokeWeight iWeight, TFontPosture iPosture, TInt aFontSize,
               FontEncoding* aEncoding);

public:
  FontEncoding* iEncoding;
};


TGfxFontSpec::TGfxFontSpec(const TDesC& aName, TFontStrokeWeight iWeight, TFontPosture iPosture, TInt aFontSize, FontEncoding* aEncoding)
  : TFontSpec(aName, aFontSize), iEncoding(aEncoding)
{
  iFontStyle.SetStrokeWeight(iWeight);
  iFontStyle.SetPosture(iPosture);
}


/////////////////////////////////////
//
// CGDIFontCache
//
// match GfxFonts with EPOC fonts
//
/////////////////////////////////////
//------------------------------------------------------------------------
// Font substitutions
//------------------------------------------------------------------------
struct TFontMapEntry {
  char*             iPdfName;
  char*             iEpocName;
  TFontStrokeWeight iWeight;
  TFontPosture      iPosture;
};

static const TInt KFontMapMax  = 14;
static const TInt KSymbol      = 12;
static const TInt KZapfDingbats= 13;

static const TFontMapEntry KFontMap[KFontMapMax] = {
  {"Courier",              "Courier New",    EStrokeWeightNormal,EPostureUpright},
  {"Courier-Bold",         "Courier New",    EStrokeWeightBold,  EPostureUpright},
  {"Courier-BoldOblique",  "Courier New",    EStrokeWeightBold,  EPostureItalic },
  {"Courier-Oblique",      "Courier New",    EStrokeWeightNormal,EPostureItalic },
  {"Helvetica",            "Arial",          EStrokeWeightNormal,EPostureUpright},
  {"Helvetica-Bold",       "Arial",          EStrokeWeightBold,  EPostureUpright},
  {"Helvetica-BoldOblique","Arial",          EStrokeWeightBold,  EPostureItalic },
  {"Helvetica-Oblique",    "Arial",          EStrokeWeightNormal,EPostureItalic },
  {"Times-Roman",          "Times New Roman",EStrokeWeightNormal,EPostureUpright},
  {"Times-Bold",           "Times New Roman",EStrokeWeightBold,  EPostureUpright},
  {"Times-Italic",         "Times New Roman",EStrokeWeightNormal,EPostureItalic },
  {"Times-BoldItalic",     "Times New Roman",EStrokeWeightBold,  EPostureItalic },
  {"Symbol",               "Psion Symbol",   EStrokeWeightNormal,EPostureUpright},
  {"ZapfDingbats",         "ZapfDingbats",   EStrokeWeightNormal,EPostureUpright}
};


struct TFontSubst {
  TFontStrokeWeight iWeight;
  TFontPosture      iPosture;
  double            iWidth;
};


// index: {symbolic:12, fixed:8, serif:4, sans-serif:0} + bold*2 + italic
static const TFontSubst KFontSubst[16] = {
  { /* Arial           */ EStrokeWeightNormal, EPostureUpright, 0.833},
  { /* Arial           */ EStrokeWeightNormal, EPostureItalic,  0.833},
  { /* Arial           */ EStrokeWeightBold  , EPostureUpright, 0.889},
  { /* Arial           */ EStrokeWeightBold  , EPostureItalic,  0.889},
  { /* Times New Roman */ EStrokeWeightNormal, EPostureUpright, 0.788},
  { /* Times New Roman */ EStrokeWeightNormal, EPostureItalic,  0.722},
  { /* Times New Roman */ EStrokeWeightBold  , EPostureUpright, 0.833},
  { /* Times New Roman */ EStrokeWeightBold  , EPostureItalic,  0.778},
  { /* Courier New     */ EStrokeWeightNormal, EPostureUpright, 0.600},
  { /* Courier New     */ EStrokeWeightNormal, EPostureItalic,  0.600},
  { /* Courier New     */ EStrokeWeightBold  , EPostureUpright, 0.600},
  { /* Courier New     */ EStrokeWeightBold  , EPostureItalic,  0.600},
  { /* Symbol Medium   */ EStrokeWeightNormal, EPostureUpright, 0.576},
  { /* Symbol Medium   */ EStrokeWeightNormal, EPostureItalic,  0.576},
  { /* Symbol Medium   */ EStrokeWeightBold  , EPostureUpright, 0.576},
  { /* Symbol Medium   */ EStrokeWeightBold  , EPostureItalic,  0.576}
};

CGDIFontCache::CGDIFontCache(CGraphicsDevice* aGd, TGDIFontNameSubst* aFontNameSubst)
  : iWindowsLatin1Encoding(*Global().Epoc.iWindowsLatin1Encoding),
    iSymbolEncoding(*Global().GfxFont.symbolEncoding),
    iZapfDingbatsEncoding(*Global().GfxFont.zapfDingbatsEncoding),
    iGd(aGd),
    iFontNameSubst(aFontNameSubst)
{}


void CGDIFontCache::ConstructL()
{
  iCache = new(ELeave) CFontCache();
}


CGDIFontCache::~CGDIFontCache()
{
  if (iCache) {
    // clear the cache
    CFont* first;
    while (0 != (first = iCache->RemoveFirstEntry()))
      iGd->ReleaseFont(first);
    delete iCache;
  }
}


void CGDIFontCache::GetGDIFont(GfxFont* aGfxFont, TReal aM[4], TInt aFontSize,
                               TGfxFontSpec& aEpocFont)
{
  TBool builtin = EFalse;
  
  TBuf8<2*KMaxTypefaceNameLength> epocName;
  TFontStrokeWeight epocWeight = EStrokeWeightNormal;
  TFontPosture      epocPosture= EPostureUpright;
  FontEncoding*     encoding   = &iWindowsLatin1Encoding;
  TBool             epocSymbol = EFalse;

  TInt i;

  if (aGfxFont->getName()) {

    // Translate font name into a epoc font name
    epocName = aGfxFont->getName()->getTPtr();
    switch (aGfxFont->getType()) {
    case fontType1:
    case fontType1C:
      // if there a '+' at the seventh char, the name is after the '+'
      if (epocName.Length() > 6 && epocName[6] == '+')
        epocName.Delete(0, 7);
      for (i = 0; i<KFontMapMax; i++) {
        if (epocName.Find(_L(KFontMap[i].iPdfName)) == 0) {
          builtin = ETrue;
          epocName   = _L(KFontMap[i].iEpocName);
          epocWeight = KFontMap[i].iWeight;
          epocPosture= KFontMap[i].iPosture;
          if (i == KSymbol) {
            encoding = &iSymbolEncoding;
            epocSymbol = ETrue;
          }
          else if (i == KZapfDingbats) {
            encoding = &iZapfDingbatsEncoding;
            epocSymbol = ETrue;
          }
          break;
        }
      }
      break;
    
    case fontTrueType:
      // if there a '+' at the seventh char, the name is after the '+'
      if (epocName.Length() > 6 && epocName[6] == '+')
        epocName.Delete(0, 7);

      // fill it the weight and posture if possible
      if (epocName.Length() > 11  && epocName.Find(_L(",BoldItalic")) != KErrNotFound) {
        epocWeight = EStrokeWeightBold;
        epocPosture= EPostureItalic;
        epocName.SetLength(epocName.Length() - 11);
      }
      else if (epocName.Length() > 5  && epocName.Find(_L(",Bold")) != KErrNotFound) {
        epocWeight = EStrokeWeightBold;
        epocName.SetLength(epocName.Length() - 5);
      }
      else if (epocName.Length() > 6  && epocName.Find(_L(",Italic")) != KErrNotFound) {
        epocPosture = EPostureItalic;
        epocName.SetLength(epocName.Length() -6);
      }
      break;
      
    default:
      break;
    }
    if (epocName.Length() > KMaxTypefaceNameLength)
      epocName.SetLength(KMaxTypefaceNameLength);
    
  }
  
  
  if (!builtin) {
    // no built-in font found, so substitute a font
    
    if (aGfxFont->is16Bit()) {
      epocName = _L(KFontMap[0].iEpocName);
      switch (aGfxFont->getCharSet16()) {
      case font16AdobeJapan12:
#if JAPANESE_SUPPORT
        xFontName = japan12Font ? japan12Font->getCString() : japan12DefFont;
#endif
        break;
      }
    } 
    else {
      if (aGfxFont->isFixedWidth())
        i = 8;
      else if (aGfxFont->isSerif())
        i = 4;
      else
        i = 0;
      if (aGfxFont->isBold()) i += 2;
      if (aGfxFont->isItalic()) i += 1;

      if (!aGfxFont->getName())
        epocName    = iFontNameSubst[i>>2].iName;
      epocWeight  = KFontSubst[i].iWeight;
      epocPosture = KFontSubst[i].iPosture;
      
      // resize the substituted font if necessary

      //TReal ntm11 = m11;
      //TReal ntm12 = m12;
      TReal ntm21 = aM[2]; // m21;
      TReal ntm22 = aM[3]; // m22;
    
      // get width of 'm' in original and substituted fonts
      TInt code;
      double w1;
      if ((code = aGfxFont->getCharCode("m")) >= 0)
        w1 = aGfxFont->getWidth((unsigned char)code);
      else
        w1 = 0;

      double w2 = KFontSubst[i].iWidth;

      if (aGfxFont->getType() == fontType3) {
        // This is a hack which makes it possible to substitute for some
        // Type 3 fonts.  The problem is that it's impossible to know what
        // the base coordinate system used in the font is without actually
        // rendering the font.  This code tries to guess by looking at the
        // width of the character 'm' (which breaks if the font is a
        // subset that doesn't contain 'm').
        if (w1 > 0 && (w1 > 1.1 * w2 || w1 < 0.9 * w2)) {
          w1 /= w2;
          //ntm11 = m11 * w1;
          //ntm12 = m12 * w1;
          ntm21 = aM[2] /*m21*/ * w1;
          ntm22 = aM[3] /*m22*/ * w1;
        }
        double* fm = aGfxFont->getFontMatrix();
        double v = (fm[0] == 0) ? 1 : (fm[3] / fm[0]);
        //ntm12 *= v;
        ntm22 *= v;
    
        // resize
        TReal size1;
        TReal size2;
        (void) Math::Sqrt(size1, ntm21*ntm21 + ntm22*ntm22);
        (void) Math::Sqrt(size2, aM[2]*aM[2] + aM[3]*aM[3]);
        aFontSize = TInt(aFontSize*size1/size2);
      } 
      else if (!aGfxFont->isSymbolic()) {
        // if real font is narrower than substituted font,
        // reduce the font size accordingly
        if (w1 > 0.01 && w1 < 0.9 * w2)
          aFontSize = TInt(aFontSize*w1/w2);
      }
      
      if (aFontSize==0) 
        aFontSize = 200; // hack hack hack: just to display something
    }
  }

  // may overrule
  if (aGfxFont->isBold())
    epocWeight = EStrokeWeightBold;
  if (aGfxFont->isItalic())
    epocPosture= EPostureItalic;

  aEpocFont = TGfxFontSpec(epocName, epocWeight, epocPosture, aFontSize, encoding);
  aEpocFont.iTypeface.SetIsProportional(!aGfxFont->isFixedWidth());
  aEpocFont.iTypeface.SetIsSerif(aGfxFont->isSerif());
  aEpocFont.iTypeface.SetIsSymbol(epocSymbol);
}


void CGDIFontCache::FillMaps(GfxFont* aGfxFont,
                             const TGfxFontSpec& aEpocFont,                           
                             /*Gushort*/ TUint aMap[],
                             Guchar  aRevMap[])
{
  // Construct forward and reverse map.
  // This tries to deal with font subset character names of the
  // form 'Bxx', 'Cxx', 'Gxx', with decimal or hex numbering.
  if (!aGfxFont->is16Bit()) {
    register TInt code, code2;
    register const char* charName;
    
    if (aEpocFont.iEncoding) {
      TBool hex = EFalse;
      for (code = 0; code < 256; ++code) {
        if ((charName = aGfxFont->getCharName(code)) != 0) {
          if ((charName[0] == 'B' || charName[0] == 'C' || charName[0] == 'G') &&
               User::StringLength((const TUint8*)charName) == 3 &&
              ((charName[1] >= 'a' && charName[1] <= 'f') ||
               (charName[1] >= 'A' && charName[1] <= 'F') ||
	       (charName[2] >= 'a' && charName[2] <= 'f') ||
	       (charName[2] >= 'A' && charName[2] <= 'F'))) {
          hex = ETrue;
          break;
          }
        }
      }
      for (code = 0; code < 256; ++code) {
        aRevMap[code] = 0;
        if ((charName = aGfxFont->getCharName(code)) != 0) {
          if ((code2 = aEpocFont.iEncoding->getCharCode(charName)) < 0) {
            TInt n = User::StringLength((const TUint8*)charName);
            if (hex && n == 3 &&
                (charName[0] == 'B' || charName[0] == 'C' || charName[0] == 'G'  /* ||
                 charName[0] == 'b' || charName[0] == 'c' || charName[0] == 'g' */) 
                && isxdigit(charName[1]) && isxdigit(charName[2])) {
              code2 = HexToInt(charName[1], charName[2]);
            } 
            else if (!hex && n >= 2 && n <= 3 && isdigit(charName[0]) && isdigit(charName[1])) {
              code2 = atoi(charName);
              if (code2 >= 256)
                code2 = -1;
            } 
            else if (!hex && n >= 3 && n <= 5 && isdigit(charName[1])) {
              code2 = atoi(charName+1);
              if (code2 >= 256)
               code2 = -1;
            }
            
            //~ this is a kludge -- is there a standard internal encoding
            //~ used by all/most Type 1 fonts?
            if (code2 == 262)		// hyphen
              code2 = 45;
            else if (code2 == 266)	// emdash
              code2 = 208;
          }
          
          if (code2 >= 0) {
            aMap[code] = (Gushort)code2;
            if (code2 < 256)
              aRevMap[code2] = (Guchar)code;
          } 
          else
            aMap[code] = 0;
        } 
        else
          aMap[code] = 0;
      }
    } 
    else { // no encoding
      code2 = 0; // to make gcc happy
      //~ this is a hack to get around the fact that X won't draw
      //~ chars 0..31; this works when the fonts have duplicate encodings
      //~ for those chars
      for (code = 0; code < 32; ++code) { 
        aRevMap[code] = 0;
        charName = aGfxFont->getCharName(code);
        if (charName && (code2 = aGfxFont->getCharCode(charName)) >= 0) {
          aMap[code] = (Gushort)code2;
          if (code2 < 256)
            aRevMap[code2] = (Guchar)code;
        }
      }
      for (code = 32; code < 256; ++code) {
        aMap[code] = (Gushort)code;
        aRevMap[code] = (Guchar)code;
      }
    }
  }
}
 

//
// translate GfxFont to a TFontSpec
//
void CGDIFontCache::Search(GfxFont* aGfxFont, TInt aFontSize, TReal aM[4], 
                           CFont*& aFont, TUint aMap[], Guchar aRevMap[])
{
  PROFILE_START(CEPOCFONTCACHE__SEARCH);

  // translate the gfx fontname into the EPOC equivalent
  TGfxFontSpec fontSpec;
  GetGDIFont(aGfxFont, aM, aFontSize, fontSpec);

  FillMaps(aGfxFont, fontSpec, aMap, aRevMap);
  
  // Translate the fontSpec to a CFont
  aFont = iCache->Search(fontSpec);
  if (!aFont) {
    User::LeaveIfError(iGd->GetNearestFontInTwips(aFont, fontSpec));
    CFont* oldFont = iCache->AddEntryL(aFont, fontSpec);
    if (oldFont)
      iGd->ReleaseFont(oldFont);
  }

  PROFILE_STOP(CEPOCFONTCACHE__SEARCH);
}


///////////////////////////////////
//
// CPolygon
//
///////////////////////////////////
CPolygon::CPolygon(TInt aGranularity): CArrayFixFlat<TPoint>(aGranularity) {}


TRect CPolygon::BoundingRect() const
{
  TRect rect(At(0), At(0));
  for (TInt i = 0; i<Count()-1; i++) {
    // bounding of rect so far and rect of segment
    rect.BoundingRect(TRect(At(i), At(i+1)));
  }
  return rect;
}


TBool CPolygon::IsRect() const
{
  if ((Count() == 5) && (At(0) == At(4))) {
    TRect rect1(At(0), At(2));
    rect1.Normalize();
    TRect rect2(At(1), At(3));
    rect2.Normalize();
    return rect1 == rect2;
  }
  else if (Count() == 4){
    TRect rect1(At(0), At(2));
    rect1.Normalize();
    TRect rect2(At(1), At(3));
    rect2.Normalize();
    return rect1 == rect2;
  }
  return EFalse;
}


// max number of splits when recursively drawing Bezier curves
static const TInt KMaxCurveSplits = 6;

void CPolygon::AddBezierL(const TPoint& aA0, const TPoint& aB0, const TPoint& aC0, const TPoint& aD0, TInt aFlatness)
{
  TInt x[(1<<KMaxCurveSplits)+1][3];
  TInt y[(1<<KMaxCurveSplits)+1][3];
  int next[1<<KMaxCurveSplits];

  // This triggers a compiler bug. (MUL rd,rm, rs and rd = rm). Now iFlatness is 
  // squared when it is set in CGDIOutputDev::updateFlatness
  // flat = iFlatness * iFlatness;
  // if (flat < 1)
  //   flat = 1;

  // initial segment
  register int p1 = 0;
  TInt p2 = 1<<KMaxCurveSplits;
  next[p1] = p2;

  // calculate with scaled integers for speed and precision
  x[p1][0] = 256*aA0.iX; y[p1][0] = 256*aA0.iY;
  x[p1][1] = 256*aB0.iX; y[p1][1] = 256*aB0.iY;
  x[p1][2] = 256*aC0.iX; y[p1][2] = 256*aC0.iY;
  x[p2][0] = 256*aD0.iX; y[p2][0] = 256*aD0.iY;

  while (p1 < (1<<KMaxCurveSplits)) {

    // get next segment
    p2 = next[p1];

    // compute distances from control points to midpoint of the
    // straight line (this is a bit of a hack, but it's much faster
    // than computing the actual distances to the line)
   
    TInt d0X = x[p2][0]; 
    TInt d0Y = y[p2][0];
    
    // if curve is flat enough, or no more divisions allowed then
    // add the straight line segment (scaled back)
    if (p2-p1<=1) {
      AppendL(TPoint(d0X/256, d0Y/256));
      p1 = p2;
    }
    else {
      TInt mx = (x[p1][0] + d0X)/2;
      TInt dx =  x[p1][1] - mx;
      TInt my = (y[p1][0] + d0Y)/2;
      TInt dy =  y[p1][1] - my;
      
      if (dx*dx + dy*dy <= aFlatness) {
        AppendL(TPoint(d0X/256, d0Y/256));
        p1 = p2;
      } 
      else {  
        dx = x[p1][2] - mx;
        dy = y[p1][2] - my;
        if (dx*dx + dy*dy <= aFlatness) {
          AppendL(TPoint(d0X/256, d0Y/256));
          p1 = p2;
        } 
        else { // otherwise, subdivide the curve with deCasteljou's algorithm
             
          register TInt p3 = (p2 + p1)/2;
          next[p1] = p3;
          
          register TInt a;
          register TInt b;
          register TInt c;
          // optimised for 'liveliness', should look at generated arm code
          b       =  x[p1][1];
          a       = (x[p1][0] + b  )/2;
          c       =  x[p1][2];
          b       = (c        + b  )/2;
          x[p1][1]=  a;
          a       = (a        + b  )/2;
          x[p1][2]=  a;
          c       = (c        + d0X)/2;
          x[p3][2]=  c;
          b       = (c        + b  )/2;
          x[p3][1]=  b;
          x[p3][0]= (b        + a  )/2;

          // liveliness analysis as for the x coordinates
          b       =  y[p1][1];
          a       = (y[p1][0] + b  )/2;
          c       =  y[p1][2];
          b       = (c        + b  )/2;
          y[p1][1]=  a;
          a       = (a        + b  )/2;
          y[p1][2]=  a;
          c       = (c        + d0Y)/2;
          y[p3][2]=  c;
          b       = (c        + b  )/2;
          y[p3][1]=  b;
          y[p3][0]= (b        + a  )/2;
          

          next[p3] = p2;
        }
      }
    }
  }
}


///////////////////////////////////
//
// CGDIView
//
///////////////////////////////////
TInt CGDIView::FindPixel(const TRgb& aColor, TDisplayMode aDisplayMode)
{
  switch (aDisplayMode) {
  case EGray2:
    return aColor.Gray2();
    break;
  case EGray4:
    return aColor.Gray4();
    break;
  case EGray16:
    return aColor.Gray16();
    break;
  case EGray256:
    return aColor.Gray256();
    break;
  case EColor16:
    return aColor.Color16();
    break;
  case EColor256:
    return aColor.Color256();
    break;
  case EColor64K:
    return ((aColor.Blue()&0xf8)<<11) | ((aColor.Green()&0xfc)<<5) | ((aColor.Red()&0xf8)>>3);
    break;
  case EColor16M: case ERgb:
    return (aColor.Blue() << 16) | (aColor.Green() << 8) | aColor.Red();
    break;
  case EColor4K:
    return ((aColor.Blue()&0xf0)<<4) | (aColor.Green()&0xf0) | ((aColor.Red()&0xf0)>>4);
    break;
  case ENone:
    break;
  }
  return 0;
}


///////////////////////////////////
//
// CGDITextView
//
///////////////////////////////////

// character substitutions

static const TInt KLastRegularChar = 0x0ff;
static const TInt KFirstSubstChar  = 0x100;
static const TInt KLastSubstChar   = 0x101;
static const TInt KFirstMultiChar  = 0x102;
static const TInt KLastMultiChar   = 0x108;
//static const TInt KFirstConstrChar = 0x109;
//static const TInt KLastConstrChar  = 0x109;

// substituted chars
static const Guchar KSubstChars[] = {
  '\x2f',      // 100: fraction --> slash
  '\xb0'       // 101: ring --> degree
};


// built-up chars
static const char* const KMultiChars[7] = {
  "fi",       // 102: fi
  "fl",       // 103: fl
  "ff",       // 104: ff
  "ffi",      // 105: ffi
  "ffl",      // 106: ffl
  "\xb4\xb4", // 107: hungarumlaut: two acute's
  "`\xb4"     // 108: caron: grave acute
};


CGDITextView::CGDITextView()
{
  iChar = TBuf<1>(1); // set the length of iChar at 1
}


CGDITextView::CGDITextView(const CGDITextView& aOriginal)
{
  iFillColor = aOriginal.iFillColor;

  iFont = aOriginal.iFont;
  iFontSizeTimesScaling = aOriginal.iFontSizeTimesScaling; 

  iChar = aOriginal.iChar;
  iPos = aOriginal.iPos;
 
  for (TInt i = 0; i < 255; i++) {
    iMap[i] = aOriginal.iMap[i];
    iRevMap[i] = aOriginal.iRevMap[i];
  }
}


CGDITextView* CGDITextView::CopyL()
{
  return new(ELeave) CGDITextView(*this);
}


TInt CGDITextView::Rotate() const { return 0; }


void CGDITextView::UpdateFillColor(const TRgb& aFillColor, CGraphicsContext* aGc)
{
  iFillColor = aFillColor;
  switch (iRender) {
    case 0: // fill text
    case 2: // fill then stroke text
    case 4: // fill text and add it to the clipping path
    case 6: // fill then stroke text and add it to the clipping path
      aGc->SetPenColor(iFillColor);
      break;
    default:
      break;
  } 
}


void CGDITextView::UpdateStrokeColor(const TRgb& aStrokeColor, CGraphicsContext* aGc)
{
  iStrokeColor = aStrokeColor;
  switch (iRender) {
    case 1: case 5:// stroke text
      aGc->SetPenColor(iStrokeColor);
      break;
    default:
      break;
  } 
}


void CGDITextView::UpdateRender(TInt aRender, CGraphicsContext* aGc)
{
  iRender = aRender;
  switch (iRender) {
    case 0: // fill text
    case 2: // fill then stroke text
    case 4: // fill text and add it to the clipping path
    case 6: // fill then stroke text and add it to the clipping path
      aGc->SetPenColor(iFillColor);
      break;

    case 1: // stroke text
    case 5: // stroke text and add it to the clipping path
      aGc->SetPenColor(iStrokeColor);
      break;

    case 3: // invisible text
    case 7: // add invisible text to the clipping path
      break;

    default:
      break;
  }
}


TInt CGDITextView::FindTextPixel(TDisplayMode aDisplayMode)
{
  switch (iRender) {
    case 0: // fill text
    case 2: // fill then stroke text
    case 4: // fill text and add it to the clipping path
    case 6: // fill then stroke text and add it to the clipping path
      return FindPixel(iFillColor, aDisplayMode);
      break;

    case 1: // stroke text
    case 5: // stroke text and add it to the clipping path
      return FindPixel(iStrokeColor, aDisplayMode);
      break;

    default:
      __ASSERT_DEBUG(EFalse, User::Invariant());
      break;
  }
  return 0;
}


void CGDITextView::UpdateFont(GfxState *aState, CGraphicsContext* aGc, CGDIFontCache* aFontCache)
{
  double m[4];
  aState->getFontTransMat(&m[0], &m[1], &m[2], &m[3]);
  m[0] *= aState->getHorizScaling();
  m[2] *= aState->getHorizScaling();

  // update the cached font size
  iFontSizeTimesScaling = aState->getFontSize()*aState->getHorizScaling();
    
  // translate scaled fontsize to user coordinates
  double fsUser, fsUserX, fsUserY;
  aState->textTransformDelta(0, iFontSizeTimesScaling, &fsUserX, &fsUserY);
   
  if (fsUserX != 0 && fsUserY == 0)
    fsUser = fsUserX;
  else if (fsUserY != 0 && fsUserX == 0)
    fsUser = fsUserY;
  else
    (void)Math::Sqrt(fsUser, fsUserX*fsUserX + fsUserY*fsUserY);

  if (fsUser < 0)
    fsUser = -fsUser;

  // Transform size back to twips, look for font with this size
  aFontCache->Search(aState->getFont(), KTwipsPerPoint*TInt(fsUser), m,
                     iFont, iMap, iRevMap);
  aGc->UseFont(iFont);
}


void CGDITextView::BeginString(GfxState* aState)
{
  SetPos(aState); 
}


void CGDITextView::DrawChar(CGraphicsContext* aGc, Guchar aChar, TInt aDx)
{
  // draw char
  TInt mappedChar = iMap[aChar];
  if (mappedChar <= KLastRegularChar) {
    iChar[0] = Guchar(mappedChar);
    aGc->DrawText(iChar, iPos);
    iPos.iX += aDx;
  }
  else 
    DrawSpecialChar(aGc, mappedChar, aDx);
}


void CGDITextView::EndString(GfxState* /* aState */)
{
}


void CGDITextView::SetPos(GfxState* aState)
{
  // calculate textpos in device space
  double dx, dy;
  double x1, y1;
  aState->textTransformDelta(0, aState->getRise(), &dx, &dy);
  aState->transform(aState->getCurX() + dx, aState->getCurY() + dy, &x1, &y1);
  iPos.iX = round(x1);
  iPos.iY = round(y1);
}


void CGDITextView::DrawString(GfxState* aState, CGraphicsContext* aGc, GString* aString)
{
  // calculate widths in user space, remap chars 
  register GfxFont* font = aState->getFont();
  
  register Guchar* p = (Guchar *)aString->getCString();
  Guchar* sub = p;
  Guchar* end = p+aString->getLength();

  TInt numGaps=0;
  register TInt numSubChars = 0; 
  register double subWidth = 0;
  
  do {
    register Guchar c = *p;
    if (c == ' ')
      numGaps++;
    TUint mappedChar = iMap[c];
    if (mappedChar <= KLastRegularChar) {
      *p = Guchar(mappedChar);
      subWidth += font->getWidth(c);
      numSubChars++;
    }
    else {
      DrawSubString(aState, aGc, sub, numSubChars, subWidth, numGaps);
      numSubChars = numGaps = 0;
      sub = p+1;
      subWidth = 0;
      double uCharWidth, uh, dx, dy; 
      aState->textTransformDelta(iFontSizeTimesScaling*aState->getFont()->getWidth(c),
                                 &uCharWidth, &uh);
      aState->transformDelta(uCharWidth, 0, &dx, &dy);
      DrawSpecialChar(aGc, mappedChar, TInt(dx));
    }
    p++;
  } while (p < end);

  if (numSubChars > 0)
    DrawSubString(aState, aGc, sub, numSubChars, subWidth, numGaps);
}


void CGDITextView::DrawSubString(GfxState* aState, CGraphicsContext* aGc, Guchar* aSubString, TInt aNumChars, double aUserWidth, TInt aNumGaps)
{
  // calculate widths in device space
  double uw, uh, x, y;
 
  // Spaces are treated separately, because you can have a large excess for
  // spaces separate from an excess in all chars.
  if (aNumGaps > 0 && aState->getWordSpace() > 0) {
    aState->textTransformDelta(aState->getWordSpace()*aNumGaps, &uw, &uh);
    aState->transformDelta(uw, uh, &x, &y);
    aGc->SetWordJustification(DevWidth(x, y), aNumGaps);
  }
  
  aState->textTransformDelta(iFontSizeTimesScaling*aUserWidth+aState->getCharSpace()*aNumChars,
                             &uw, &uh);
  aState->transformDelta(uw, uh, &x, &y);
  TInt devCharsWidth = DevWidth(x, y);

  TPtrC8 string((TUint8*)aSubString, aNumChars);
  
  // The excess is the mismatch between the PDF font width and the device font width
  aGc->SetCharJustification(devCharsWidth - iFont->TextWidthInPixels(string), aNumChars);

  // render the text
  aGc->DrawText(string, iPos);
  iPos.iX += devCharsWidth;
}


void CGDITextView::DrawSpecialChar(CGraphicsContext* aGc, TInt aMappedChar, TInt aDx)
{
  if (aMappedChar <= KLastSubstChar) {

    // substitute char
    iChar[0] = Guchar(KSubstChars[aMappedChar - KFirstSubstChar]);

    aGc->DrawText(iChar, iPos);
  }
  else if (aMappedChar <= KLastMultiChar) {

    // draw the multichar in the space reserved for the mapped char
    TPtrC multiChar = (TUint8*)KMultiChars[aMappedChar - KFirstMultiChar];

    aGc->SetCharJustification(aDx - iFont->TextWidthInPixels(multiChar), 
                              multiChar.Length());
    aGc->DrawText(multiChar, iPos);
  }
  else { // now what, can't draw it
    aGc->DrawText(_L("?"), iPos);
  }
  
  iPos.iX += aDx;
}


TInt CGDITextView::DevWidth(const TReal& aX, const TReal& /* aY */)
{
  return TInt(aX);
}


///////////////////////////////////
//
// CRotatedTextView
//
///////////////////////////////////
CRotatedTextView::CRotatedTextView(const CRotatedTextView& aOriginal)
  : CGDITextView(aOriginal),
    iRotate(aOriginal.iRotate),
    iRotDev(aOriginal.iRotDev),
    iFontNameSubst(aOriginal.iFontNameSubst),
    iZeroMap(aOriginal.iZeroMap),
    iZeroDev(aOriginal.iZeroDev),
    iZeroCache(aOriginal.iZeroCache),
    iZeroGc(aOriginal.iZeroGc),
    iRotStart(aOriginal.iRotStart),
    iZeroStart(aOriginal.iZeroStart)
{}


CRotatedTextView::CRotatedTextView(const CGDITextView& aOriginal, TInt aRotate, CGraphicsDevice* aRotDev, TGDIFontNameSubst* aFontNameSubst)
  : CGDITextView(aOriginal),
    iRotate(aRotate),
    iRotDev(aRotDev),
    iFontNameSubst(aFontNameSubst)
{}


void CRotatedTextView::ConstructL()
{ 
  SetZeroSizes(20);
 
  // create the unrotated bitmap
  iZeroMap = new(ELeave) CWsBitmap(CCoeEnv::Static()->WsSession());
  User::LeaveIfError(iZeroMap->Create(iZeroSizeInPixels, EGray2));
  
  iZeroMap->SetSizeInTwips(iZeroSizeInTwips);
      
  iZeroDev = CFbsBitmapDevice::NewL(iZeroMap);

  iZeroCache = new(ELeave) CGDIFontCache(iZeroDev, iFontNameSubst);
  iZeroCache->ConstructL();

  User::LeaveIfError(iZeroDev->CreateContext(iZeroGc));
}


CRotatedTextView::~CRotatedTextView()
{
  delete iZeroGc;
  delete iZeroCache;
  delete iZeroDev;
  delete iZeroMap; 
}


TInt CRotatedTextView::Rotate() const { return iRotate; }


void CRotatedTextView::UpdateFont(GfxState *aState, CGraphicsContext* /*aGc*/, CGDIFontCache* /*aFontCache*/)
{ 
  CGDITextView::UpdateFont(aState, iZeroGc, iZeroCache);
  TInt newHeight = iFont->HeightInPixels();
  if (newHeight > iZeroSizeInPixels.iHeight) {
  
    //resize the bitmap
    SetZeroSizes(newHeight);
    
    User::LeaveIfError(iZeroMap->Resize(iZeroSizeInPixels));
    iZeroMap->SetSizeInTwips(iZeroSizeInTwips);
    
    User::LeaveIfError(iZeroDev->Resize(iZeroSizeInPixels));
    iZeroGc->Resized();
  }
}


void CRotatedTextView::BeginString(GfxState* aState)
{
  CGDITextView::BeginString(aState);
  
  // set iPos to the start position in the rotated device space
  iRotStart = iPos;

  // now set iPos to the start position in the unrotated device space
  iPos = UnrotatePos(iRotStart);
  iZeroStart = iPos;
  iZeroStart.iY -= iFont->AscentInPixels();
}


void CRotatedTextView::DrawChar(CGraphicsContext* /*aGc*/, Guchar aChar, TInt aDx)
{
  iPos = UnrotatePos(iPos);
  CGDITextView::DrawChar(iZeroGc, aChar, aDx);
}


void CRotatedTextView::EndString(GfxState* aState)
{
  CGDITextView::EndString(aState);
  
  // calculate rotated textpos in device space
  CGDITextView::SetPos(aState);
  iPos = UnrotatePos(iPos);

  // now: bitblit the drawn string
  iPos.iY += iFont->DescentInPixels();
  
  TBitmapUtil* util0 = new(ELeave) TBitmapUtil(iZeroMap);
  CleanupStack::PushL(util0);
  util0->Begin(iZeroStart);
  
  DrawRotatedStringL(util0, iPos);
  
  util0->End();
  CleanupStack::PopAndDestroy(); // util0;

  iZeroGc->Clear();
}


void CRotatedTextView::DrawString(GfxState* aState, CGraphicsContext* /*aGc*/, GString* aString)
{
  // set iPos to the start position in the rotated device space
  iRotStart = iPos;
  
  // now set iPos to the start position in the unrotated device space
  iPos = UnrotatePos(iRotStart);
  iZeroStart = iPos;
  iZeroStart.iY -= iFont->AscentInPixels();
 
  // draw to the unrotated device, updating iPos in the unrotated device space
  CGDITextView::DrawString(aState, iZeroGc, aString);
  
  // copy the unrotated string to the rotated device
  iPos.iY += iFont->DescentInPixels();
  
  TBitmapUtil* util0 = new(ELeave) TBitmapUtil(iZeroMap);
  CleanupStack::PushL(util0);
  util0->Begin(iZeroStart);
  
  DrawRotatedStringL(util0, iPos);
  
  util0->End();
  CleanupStack::PopAndDestroy(); // util0

  // clear the unrotated device
  iZeroGc->Clear();
}

 
void CRotatedTextView::SetZeroSizes(TInt aNewHeight)
{
  iZeroSizeInPixels.iHeight = aNewHeight;
  switch (iRotate) {
  case 180:
    iZeroSizeInPixels.iWidth= iRotDev->SizeInPixels().iWidth;
    
    iZeroSizeInTwips.iWidth = iRotDev->HorizontalPixelsToTwips(iZeroSizeInPixels.iWidth);
    iZeroSizeInTwips.iHeight= iRotDev->VerticalPixelsToTwips(iZeroSizeInPixels.iHeight);
    break;

  case 90:
  case 270:
    iZeroSizeInPixels.iWidth= iRotDev->SizeInPixels().iHeight;

    iZeroSizeInTwips.iWidth = iRotDev->VerticalPixelsToTwips(iZeroSizeInPixels.iWidth);
    iZeroSizeInTwips.iHeight= iRotDev->HorizontalPixelsToTwips(iZeroSizeInPixels.iHeight);
    break;

  default:
    __ASSERT_ALWAYS(EFalse, User::Invariant());
    break;
  }
}


TPoint CRotatedTextView::UnrotatePos(const TPoint aPos) const
{
  switch (iRotate) {
  case 90:
    return TPoint(aPos.iY, iFont->AscentInPixels());
    break;
  case 180:
    return TPoint(iZeroMap->SizeInPixels().iWidth  - aPos.iX, iFont->AscentInPixels());
    break;
  case 270:
    return TPoint(iZeroMap->SizeInPixels().iWidth  - aPos.iY, iFont->AscentInPixels());
    break;
  default:
    __ASSERT_DEBUG(EFalse, User::Invariant());
    break;
  }
  return TPoint (0,0);;
}


TInt CRotatedTextView::DevWidth(const TReal& aX, const TReal& aY)
{
  switch (iRotate) {
  case 90:
    return TInt(aY);
  case 180:
    return TInt(-aX);
  case 270:
    return TInt(-aY);
  default:
    __ASSERT_DEBUG(EFalse, User::Invariant());
    break;
  }
  return TInt(aX);
}
